/* Copyright (C) 2016-2021 |Meso|Star> (contact@meso-star.com)
 *
 * This software is governed by the CeCILL license under French law and
 * abiding by the rules of distribution of free software. You can use,
 * modify and/or redistribute the software under the terms of the CeCILL
 * license as circulated by CEA, CNRS and INRIA at the following URL
 * "http://www.cecill.info".
 *
 * As a counterpart to the access to the source code and rights to copy,
 * modify and redistribute granted by the license, users are provided only
 * with a limited warranty and the software's author, the holder of the
 * economic rights, and the successive licensors have only limited
 * liability.
 *
 * In this respect, the user's attention is drawn to the risks associated
 * with loading, using, modifying and/or developing or reproducing the
 * software by the user in light of its specific status of free software,
 * that may mean that it is complicated to manipulate, and that also
 * therefore means that it is reserved for developers and experienced
 * professionals having in-depth computer knowledge. Users are therefore
 * encouraged to load and test the software's suitability as regards their
 * requirements in conditions enabling the security of their systems and/or
 * data to be ensured and, more generally, to use and operate it in the
 * same conditions as regards security.
 *
 * The fact that you are presently reading this means that you have had
 * knowledge of the CeCILL license and that you accept its terms. */

#include "s2d.h"
#include "s2d_device_c.h"
#include "s2d_scene_c.h"
#include "s2d_scene_view_c.h"
#include "s2d_shape_c.h"

#include <rsys/list.h>
#include <rsys/mem_allocator.h>

/*******************************************************************************
 * Helper functions
 ******************************************************************************/
static void
scene_release(ref_T* ref)
{
  struct s2d_scene* scn;
  struct s2d_device* dev;
  struct list_node* node;
  struct list_node* tmp;

  ASSERT(ref);
  scn = CONTAINER_OF(ref, struct s2d_scene, ref);
  dev = scn->dev;

  LIST_FOR_EACH_SAFE(node, tmp, &scn->scnviews) {
    scene_view_destroy(CONTAINER_OF(node, struct s2d_scene_view, node));
  }
  S2D(scene_clear(scn));
  htable_shape_release(&scn->shapes);
  MEM_RM(dev->allocator, scn);
  S2D(device_ref_put(dev));
}

/*******************************************************************************
 * Exported s2d_scene functions
 ******************************************************************************/
res_T
s2d_scene_create(struct s2d_device* dev, struct s2d_scene** out_scn)
{
  struct s2d_scene* scn = NULL;
  res_T res = RES_OK;

  if(!dev || !out_scn) {
    res = RES_BAD_ARG;
    goto error;
  }

  scn = (struct s2d_scene*)MEM_CALLOC
    (dev->allocator, 1, sizeof(struct s2d_scene));
  if(!scn) {
    res = RES_MEM_ERR;
    goto error;
  }
  htable_shape_init(dev->allocator, &scn->shapes);
  SIG_INIT(&scn->sig_shape_detach);
  list_init(&scn->scnviews);
  ref_init(&scn->ref);
  S2D(device_ref_get(dev));
  scn->dev = dev;

exit:
  if(out_scn) *out_scn = scn;
  return res;
error:
  if(scn) {
    S2D(scene_ref_put(scn));
    scn = NULL;
  }
  goto exit;
}

res_T
s2d_scene_ref_get(struct s2d_scene* scn)
{
  if(!scn) return RES_BAD_ARG;
  ref_get(&scn->ref);
  return RES_OK;
}

res_T
s2d_scene_ref_put(struct s2d_scene* scn)
{
  if(!scn) return RES_BAD_ARG;
  ref_put(&scn->ref, scene_release);
  return RES_OK;
}

res_T
s2d_scene_attach_shape(struct s2d_scene* scn, struct s2d_shape* shape)
{
  unsigned shape_id;
  res_T res = RES_OK;

  if(!scn || !shape) {
    return RES_BAD_ARG;
  }

  S2D(shape_get_id(shape, &shape_id));
  if(htable_shape_find(&scn->shapes, &shape_id) != NULL) {
    log_warning(scn->dev,
      "%s: the shape is already attached to the scene.\n", FUNC_NAME);
    return RES_OK;
  }

  res = htable_shape_set(&scn->shapes, &shape_id, &shape);
  if(res != RES_OK) {
    log_error(scn->dev,
      "%s: cannot attach the shape to the scene.\n", FUNC_NAME);
    return RES_OK;
  }
  S2D(shape_ref_get(shape));
  return RES_OK;
}

res_T
s2d_scene_detach_shape(struct s2d_scene* scn, struct s2d_shape* shape)
{
  size_t n;
  unsigned shape_id;

  if(!scn || !shape) return RES_BAD_ARG;

  S2D(shape_get_id(shape, &shape_id));
  if(htable_shape_find(&scn->shapes, &shape_id) == NULL) {
    log_error(scn->dev,
      "%s: the shape is not attached to the scene.\n", FUNC_NAME);
    return RES_BAD_ARG;
  }

  n = htable_shape_erase(&scn->shapes, &shape_id);
  ASSERT(n == 1); (void)n;

  SIG_BROADCAST(&scn->sig_shape_detach, scene_shape_cb_T, ARG2(scn, shape));

  S2D(shape_ref_put(shape));
  return RES_OK;
}

res_T
s2d_scene_clear(struct s2d_scene* scn)
{
  struct htable_shape_iterator it, end;

  if(!scn) return RES_BAD_ARG;

  htable_shape_begin(&scn->shapes, &it);
  htable_shape_end(&scn->shapes, &end);
  while(!htable_shape_iterator_eq(&it, &end)) {
    struct s2d_shape** pshape = htable_shape_iterator_data_get(&it);
    struct s2d_shape* shape = *pshape;
    SIG_BROADCAST(&scn->sig_shape_detach, scene_shape_cb_T, ARG2(scn, shape));
    S2D(shape_ref_put(shape));
    htable_shape_iterator_next(&it);
  }
  htable_shape_clear(&scn->shapes);

  return RES_OK;
}

res_T
s2d_scene_get_device(struct s2d_scene* scn, struct s2d_device** dev)
{
  if(!scn || !dev) return RES_BAD_ARG;
  *dev = scn->dev;
  return RES_OK;
}

res_T
s2d_scene_get_shapes_count(struct s2d_scene* scn, size_t* nshapes)
{
  if(!scn || !nshapes) return RES_BAD_ARG;
  *nshapes = htable_shape_size_get(&scn->shapes);
  return RES_OK;
}

