/* Copyright (C) 2020, 2021 |Meso|Star> (contact@meso-star.com)
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. If not, see <http://www.gnu.org/licenses/>. */

#ifndef GREEN_TYPES_H
#define GREEN_TYPES_H

#include <rsys/str.h>
#include <rsys/mem_allocator.h>
#include <rsys/hash_table.h>
#include <rsys/dynamic_array.h>
#include <rsys/dynamic_array_char.h>
#include <rsys/dynamic_array_str.h>

#include <string.h>
#include <omp.h>

/* Utility macros */
#define ERR(Expr) if((res = (Expr)) != RES_OK) goto error; else (void)0

#define CHK_TOK(x, Name) if((tk = (x)) == NULL) {\
   logger_print(green->logger, LOG_ERROR,\
     "Invalid data (missing " Name ")\n");\
   res = RES_BAD_ARG;\
   goto error;\
  } (void)0

#ifdef COMPILER_CL
#define strtok_r strtok_s
#endif

static INLINE void
log_err(const char* msg, void* ctx)
{
  ASSERT(msg);
  (void)ctx;
#ifdef OS_WINDOWS
  fprintf(stderr, "error: %s", msg);
#else
  fprintf(stderr, "\x1b[31merror:\x1b[0m %s", msg);
#endif
}

static INLINE void
log_warn(const char* msg, void* ctx)
{
  ASSERT(msg);
  (void)ctx;
#ifdef OS_WINDOWS
  fprintf(stderr, "warning: %s", msg);
#else
  fprintf(stderr, "\x1b[33mwarning:\x1b[0m %s", msg);
#endif
}

static INLINE void
log_prt(const char* msg, void* ctx)
{
  ASSERT(msg);
  (void)ctx;
#ifdef OS_WINDOWS
  fprintf(stderr, "message: %s", msg);
#else
  fprintf(stderr, "\x1b[32moutput:\x1b[0m %s", msg);
#endif
}

/*
 * Counts read from the binary file
 */
struct bfile_green_counts {
  unsigned desc_count, smed_count, fmed_count, tbound_count, hbound_count,
    fbound_count, sfconnect_count, ssconnect_count, name_pool_sz;
  size_t ok_count, failed_count;
};

static INLINE void
init_green_counts
  (struct bfile_green_counts* counts)
{
  ASSERT(counts);
  counts->desc_count = 0;
  counts->smed_count = 0;
  counts->fmed_count = 0;
  counts->tbound_count = 0;
  counts->hbound_count = 0;
  counts->fbound_count = 0;
  counts->sfconnect_count = 0;
  counts->ssconnect_count = 0;
  counts->name_pool_sz = 0;
  counts->ok_count = 0;
  counts->failed_count = 0;
}

/*
 * A type for green mode
 */
enum green_mode {
  MODE_NULL = 0,
  MODE_APPLY_GREEN = BIT(1), /* -a */
  MODE_EXTENTED_RESULTS = BIT(2), /* -e */
  MODE_READ_GREEN = BIT(3), /* -g */
  MODE_HELP = BIT(4), /* -h */
  MODE_HTML_SUMMARY = BIT(5), /* -s */
  MODE_THREADS_COUNT = BIT(6), /* -v */
  MODE_VERSION = BIT(7), /* -v */
  MODE_VERBOSE = BIT(8) /* -v */
};

/*
 * The type to store a green sample
 */
struct sample_header {
  unsigned end_id;
  unsigned pw_count, fx_count;
  char at_initial;
};

struct sample {
  struct sample_header header;
  unsigned* pw_ids;
  double* pw_weights;
  unsigned* fx_ids;
  double* fx_weights;
};

static INLINE res_T
alloc_sample_buffers
  (struct mem_allocator* alloc, 
   struct sample* sample)
{
  unsigned* ids = NULL;
  double* weights = NULL;
  size_t count;

  ASSERT(alloc && sample);
  count =
    sample->header.pw_count + sample->header.fx_count;
  ids = MEM_ALLOC(alloc, count * sizeof(*ids));
  weights = MEM_ALLOC(alloc, count * sizeof(*sample->pw_weights));

  if(!ids || !weights) {
    MEM_RM(alloc, ids);
    MEM_RM(alloc, weights);
    return RES_MEM_ERR;
  }
  sample->pw_ids = ids;
  sample->pw_weights = weights;
  sample->fx_ids = ids + sample->header.pw_count;
  sample->fx_weights = weights + sample->header.pw_count;

  return RES_OK;
}

#define READ_SAMPLE_BUFFER(Sample) { \
  size_t count = (Sample)->header.pw_count + (Sample)->header.fx_count; \
  FR((Sample)->pw_ids, count); \
  FR((Sample)->pw_weights, count); \
} (void)0

static INLINE void
release_sample
  (struct mem_allocator* alloc,
   struct sample* sample)
{
  ASSERT(alloc && sample);
  MEM_RM(alloc, sample->pw_ids);
  MEM_RM(alloc, sample->pw_weights);
  /* Don't RM fx_ids nor fx_weights as they share pw buffers */
}

/*
 * Types used to compute green
 */

struct applied_settings_elt {
  double imposed_T_value, initial_T_value;
  double other_value;
};

/* other is used both for power and flux, as no description can have both */
struct table_elt {
  struct str imposed_T_name, initial_T_name, other_name;
  double imposed_T_value, initial_T_value, other_value;
  int imposed_T_defined, initial_T_defined, other_defined;
  int imposed_T_unknown, initial_T_unknown, other_unknown;
  unsigned imposed_T_weight, initial_T_weight;
  double other_weight;
};

static INLINE void
init_table_elt
  (struct mem_allocator* alloc,
   struct table_elt* elt)
{
  ASSERT(alloc && elt);
  str_init(alloc, &elt->imposed_T_name);
  str_init(alloc, &elt->initial_T_name);
  str_init(alloc, &elt->other_name);
  elt->imposed_T_value = 0;
  elt->initial_T_value = 0;
  elt->other_value = 0;
  elt->imposed_T_defined = 0;
  elt->initial_T_defined = 0;
  elt->other_defined = 0;
  elt->imposed_T_unknown = 0;
  elt->initial_T_unknown = 0;
  elt->other_unknown = 0;
  elt->imposed_T_weight = 0;
  elt->initial_T_weight = 0;
  elt->other_weight = 0;
}

static INLINE void
release_table_elt
  (struct table_elt* elt)
{
  ASSERT(elt);
  str_release(&elt->imposed_T_name);
  str_release(&elt->initial_T_name);
  str_release(&elt->other_name);
}

struct variable_data {
  size_t idx;
  int used;
  int unknown;
};

#define HTABLE_NAME variable_ptr
#define HTABLE_DATA struct variable_data
#define HTABLE_KEY struct str
#define HTABLE_KEY_FUNCTOR_HASH str_hash
#define HTABLE_KEY_FUNCTOR_EQ str_eq
#define HTABLE_KEY_FUNCTOR_INIT str_init
#define HTABLE_KEY_FUNCTOR_COPY str_copy
#define HTABLE_KEY_FUNCTOR_RELEASE str_release
#define HTABLE_KEY_FUNCTOR_COPY_AND_RELEASE str_copy_and_release
#include <rsys/hash_table.h>

struct result {
  double E, STD;
};

#define DARRAY_NAME result
#define DARRAY_DATA struct result
#include <rsys/dynamic_array.h>

/*
 * Main type to store binary file content
 */
struct green {
  struct mem_allocator* allocator;
  struct logger* logger;
  struct bfile_green_counts counts;
  char* names_pool;
  struct description* descriptions;
  double ambient_temp, ref_temp;
  double time_range[2];
  struct table_elt* table;
  struct sample* samples;
  int references_checked, unused_variables, unknown_variables;
  unsigned nthreads;
  struct htable_variable_ptr variable_ptrs;
  struct darray_str settings;
  struct darray_result results;
};

static INLINE void
green_init
  (struct mem_allocator* alloc,
   struct logger* logger,
   const unsigned nthreads,
   struct green* green)
{
  ASSERT(alloc && green);
  green->allocator = alloc;
  green->logger = logger;
  green->names_pool = NULL;
  green->descriptions = NULL;
  green->table = NULL;
  green->samples = NULL;
  init_green_counts(&green->counts);
  green->ambient_temp = 0;
  green->ref_temp = 0;
  green->references_checked = 0;
  green->unused_variables = 0;
  green->unknown_variables = 0;
  green->nthreads = MMIN(nthreads, (unsigned)omp_get_num_procs());
  htable_variable_ptr_init(alloc, &green->variable_ptrs);
  darray_str_init(alloc, &green->settings);
  darray_result_init(alloc, &green->results);
}

static INLINE void
green_release
  (struct green* green)
{
  unsigned i;
  ASSERT(green);

  MEM_RM(green->allocator, green->names_pool);
  MEM_RM(green->allocator, green->descriptions);
  if(green->samples) { /* On error, samples could be not yet allocated */
    FOR_EACH(i, 0, green->counts.ok_count)
      release_sample(green->allocator, green->samples + i);
    MEM_RM(green->allocator, green->samples);
  }
  if(green->table) { /* On error, table could be not yet allocated */
    FOR_EACH(i, 0, 1 + green->counts.desc_count)
      if(&green->table[i]) release_table_elt(&green->table[i]);
    MEM_RM(green->allocator, green->table);
  }
  green->allocator = NULL;
  htable_variable_ptr_release(&green->variable_ptrs);
  darray_str_release(&green->settings);
  darray_result_release(&green->results);
}

res_T
read_green_function
  (struct green* green,
   FILE* stream);

res_T
green_read_settings
  (struct green* green,
   const char* in_name);

/*
 * Different types of descriptions read from the binary file
 */
enum description_type {
  DESC_MAT_SOLID,
  DESC_MAT_FLUID,
  DESC_BOUND_H_FOR_FLUID,
  DESC_BOUND_H_FOR_SOLID,
  DESC_BOUND_T_FOR_SOLID,
  DESC_BOUND_F_FOR_SOLID,
  DESC_SOLID_FLUID_CONNECT,
  DESC_SOLID_SOLID_CONNECT,
  DESCRIPTION_TYPE_COUNT__,
  DESC_OUTSIDE
};

#define DESC_IS_MEDIUM(D) \
  ((D) == DESC_MAT_SOLID || (D) == DESC_MAT_FLUID)
#define DESC_IS_H(D) \
  ((D) == DESC_BOUND_H_FOR_SOLID || (D) == DESC_BOUND_H_FOR_FLUID)
#define DESC_IS_T(D) \
  ((D) == DESC_BOUND_T_FOR_SOLID)
#define DESC_IS_F(D) \
  ((D) == DESC_BOUND_F_FOR_SOLID)
#define DESC_IS_SF(D) \
  ((D) == DESC_SOLID_FLUID_CONNECT)
#define DESC_IS_SS(D) \
  ((D) == DESC_SOLID_SOLID_CONNECT)
#define DESC_IS_CONNECTION(D) \
  (DESC_IS_SS(D) || DESC_IS_SF(D))
#define DESC_HOLDS_T(D) \
  (DESC_IS_MEDIUM(D) || DESC_IS_H(D) || DESC_IS_T(D))

struct mat_fluid {
  struct str name;
  double rho; /* Volumic mass */
  double cp; /* Calorific capacity */
  double tinit;
  double imposed_temperature;
  double t0; /* End time of tinit */
  int is_outside; /* the fluid is used for a boundary */
  int is_green; /* green computation nothing to do with fluid itself) */
  unsigned desc_id; /* id of the boundary; meaningful if is_outside */
  unsigned fluid_id;
};

struct mat_solid {
  struct str name;
  double lambda; /* Conductivity */
  double rho; /* Volumic mass */
  double cp; /* Calorific capacity */
  double delta; /* Numerical parameter */
  double tinit; /* Initial temperature */
  double imposed_temperature; /* allow to impose a T; -1 if unset */
  double vpower;
  double t0; /* End time of tinit */
  int is_outside; /* the fluid is used for a boundary */
  int is_green; /* green computation nothing to do with fluid itself) */
  unsigned desc_id; /* id of the boundary; meaningful if is_outside */
  unsigned solid_id;
};

struct h_boundary {
  struct str name;
  double emissivity;
  double specular_fraction;
  double hc;
  double imposed_temperature;
  unsigned mat_id;
};

struct t_boundary {
  struct str name;
  double imposed_temperature;
  unsigned mat_id;
};

struct f_boundary {
  struct str name;
  double imposed_flux;
  unsigned mat_id;
};

struct solid_fluid_connect {
  struct str name;
  double emissivity;
  double specular_fraction;
  double hc;
  unsigned connection_id;
};

struct solid_solid_connect {
  struct str name;
  double tcr;
  unsigned connection_id;
};

struct description {
  enum description_type type;
  union {
    struct mat_fluid fluid;
    struct mat_solid solid;
    struct t_boundary t_boundary;
    struct f_boundary f_boundary;
    struct h_boundary h_boundary;
    struct solid_fluid_connect sf_connect;
    struct solid_solid_connect ss_connect;
  } d;
};

static INLINE res_T
get_description_name
  (struct description* desc,
   struct str** name)
{
  ASSERT(desc && name);
  switch (desc->type) {
  case DESC_MAT_SOLID:
    *name = &desc->d.solid.name;
    break;
  case DESC_MAT_FLUID:
    *name = &desc->d.fluid.name;
    break;
  case DESC_BOUND_T_FOR_SOLID:
    *name = &desc->d.t_boundary.name;
    break;
  case DESC_BOUND_H_FOR_SOLID:
  case DESC_BOUND_H_FOR_FLUID:
    *name = &desc->d.h_boundary.name;
    break;
  case DESC_BOUND_F_FOR_SOLID:
    *name = &desc->d.f_boundary.name;
    break;
  case DESC_SOLID_FLUID_CONNECT:
    *name = &desc->d.sf_connect.name;
    break;
  case DESC_SOLID_SOLID_CONNECT:
    *name = &desc->d.ss_connect.name;
    break;
  default:
    return RES_BAD_ARG;
  }
  return RES_OK;
}

#endif /* GREEN_TYPES_H */
